@@ -1,4 +1,4 @@
-# Copyright (c) 2009, 2010, 2011, 2012, 2013 Nicira, Inc.
+# Copyright (c) 2009, 2010, 2011, 2012, 2013, 2016 Nicira, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -144,7 +144,7 @@ class Idl(object):
             table.need_table = False
             table.rows = {}
             table.idl = self
-            table.condition = []
+            table.condition = [True]
             table.cond_changed = False
 
     def close(self):
@@ -265,23 +265,23 @@ class Idl(object):
                 self.__send_cond_change(table, table.condition)
                 table.cond_changed = False
 
-    def cond_change(self, table_name, add_cmd, cond):
-        """Change conditions for this IDL session. If session is not already
-        connected, add condtion to table and submit it on send_monitor_request.
-        Otherwise  send monitor_cond_change method with the requested
-        changes."""
+    def cond_change(self, table_name, cond):
+        """Sets the condition for 'table_name' to 'cond', which should be a
+        conditional expression suitable for use directly in the OVSDB
+        protocol, with the exception that the empty condition []
+        matches no rows (instead of matching every row).  That is, []
+        is equivalent to [False], not to [True].
+        """
 
         table = self.tables.get(table_name)
         if not table:
             raise error.Error('Unknown table "%s"' % table_name)
 
-        if add_cmd:
-            table.condition += cond
-        else:
-            for c in cond:
-                table.condition.remove(c)
-
-        table.cond_changed = True
+        if cond == []:
+            cond = [False]
+        if table.condition != cond:
+            table.condition = cond
+            table.cond_changed = True
 
     def wait(self, poller):
         """Arranges for poller.block() to wake up when self.run() has something
@@ -420,8 +420,7 @@ class Idl(object):
                         (column not in self.readonly[table.name])):
                     columns.append(column)
             monitor_requests[table.name] = {"columns": columns}
-            if method == "monitor_cond" and table.cond_changed and \
-                   table.condition:
+            if method == "monitor_cond" and table.condition != [True]:
                 monitor_requests[table.name]["where"] = table.condition
                 table.cond_change = False
 
@@ -650,6 +649,7 @@ class Idl(object):
         txn = self._outstanding_txns.pop(msg.id, None)
         if txn:
             txn._process_reply(msg)
+            return True
 
 
 def _uuid_to_row(atom, base):
@@ -770,6 +770,7 @@ class Row(object):
         assert self._changes is not None
         assert self._mutations is not None
 
+        column = self._table.columns[column_name]
         datum = self._changes.get(column_name)
         inserts = None
         if '_inserts' in self._mutations.keys():
@@ -784,24 +785,33 @@ class Row(object):
                                          (self.__class__.__name__,
                                           column_name))
                 else:
-                    datum = inserts
-            if column_name in self._data:
+                    datum = data.Datum.from_python(column.type,
+                                                   inserts,
+                                                   _row_to_uuid)
+            elif column_name in self._data:
                 datum = self._data[column_name]
-                try:
+                if column.type.is_set():
+                    dlist = datum.as_list()
                     if inserts is not None:
-                        datum.extend(inserts)
+                        dlist.extend(list(inserts))
                     if removes is not None:
-                        datum = [x for x in datum if x not in removes]
-                except error.Error:
-                    pass
-                try:
+                        removes_datum = data.Datum.from_python(column.type,
+                                                              removes,
+                                                              _row_to_uuid)
+                        removes_list = removes_datum.as_list()
+                        dlist = [x for x in dlist if x not in removes_list]
+                    datum = data.Datum.from_python(column.type, dlist,
+                                                   _row_to_uuid)
+                elif column.type.is_map():
+                    dmap = datum.to_python(_uuid_to_row)
                     if inserts is not None:
-                        datum.merge(inserts)
+                        dmap.update(inserts)
                     if removes is not None:
-                        for key in removes.keys():
-                            del datum[key]
-                except error.Error:
-                    pass
+                        for key in removes:
+                            if key not in (inserts or {}):
+                                del dmap[key]
+                    datum = data.Datum.from_python(column.type, dmap,
+                                                   _row_to_uuid)
             else:
                 if inserts is None:
                     raise AttributeError("%s instance has no attribute '%s'" %
@@ -870,7 +880,7 @@ class Row(object):
             vlog.err("attempting to write bad value to column %s (%s)"
                      % (column_name, e))
             return
-        if column_name in self._data:
+        if self._data and column_name in self._data:
             # Remove existing key/value before updating.
             removes = self._mutations.setdefault('_removes', {})
             column_value = removes.setdefault(column_name, set())
